#!/usr/bin/perl

# createcatalogs.pl, version 1.1, 2012-05-14.
#
#
# Copyright (C) 2012 Andrew Tribick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Build the Celestia catalog files from the data returned from the KIC query
# and the Kepler candidates data table.

use strict;

use HTTP::Request;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError);
use LWP::UserAgent;
use Math::Trig;
use Math::Quaternion;
use Term::ProgressBar;

use constant PARSEC_LY => 3.26163626;
use constant GRAVITATION => 2.97596876e+15; # km^3 m_earth^-1 day^-2
use constant AU_KM => 149598000;

use constant SOLAR_BOLO_MAG => 4.75;
use constant SOLAR_TEMPERATURE => 5778;
use constant SOLAR_RADIUS_AU => 0.00464912633;
use constant SOLAR_RADIUS_KM => 696000;

use constant EARTH_OBLIQUITY => 23.4392911;
use constant EARTH_YEAR => 365.25;
use constant EARTH_RADIUS_KM => 6378.1;

# Some information for generating file headers
my $author = 'Andrew Tribick';

my ($day, $month, $year) = (localtime)[3,4,5];
my $date = sprintf('%04d-%02d-%02d', $year+1900, $month+1, $day);

# Get files from internet if necessary...

unless((-e 'kic.txt.gz') && (-e 'koiTable_2012Feb26.txt'))
{
  my $ua = LWP::UserAgent->new();
  $ua->env_proxy;

  unless(-e 'kic.txt.gz')
  {
    print "kic.txt.gz not found.\n";
    print "Download file? (761 MB) (y/N): ";
    my $reply = <STDIN>;
    chomp $reply;
    if ($reply =~ m/^[Yy]$/)
    {
      if (!download_file($ua, 'http://archive.stsci.edu/pub/kepler/catalogs/kic.txt.gz', 'kic.txt.gz'))
      {
        die "Exiting.\n";
      }
    }
    else
    {
      print "Exiting.\n";
      exit 0;
    }
  }

  unless(-e 'koiTable_2012Feb26.txt')
  {
    print "koiTable_2012Feb26.txt not found.\n";
    print "Download file? (456 KB) (y/N): ";
    my $reply = <STDIN>;
    chomp $reply;
    if ($reply =~ m/^[Yy]$/)
    {
      if (!download_file($ua, 'http://archive.stsci.edu/pub/kepler/catalogs/koiTable_2012Feb26.txt', 'koiTable_2012Feb26.txt'))
      {
        die "Exiting.\n";
      }
    }
    else
    {
      print "Exiting.\n";
      exit 0;
    }
  }
}

# The following stars are in Celestia so exclude them from the output.
my %exclude = (
  11446443 => 'TrES-2',
  10666592 => 'HAT-P-7',
  10748390 => 'HAT-P-11',
  11853905 => 'Kepler-4',
  8191672 => 'Kepler-5',
  10874614 => 'Kepler-6',
  5780885 => 'Kepler-7',
  6922244 => 'Kepler-8',
  3323887 => 'Kepler-9',
  11904151 => 'Kepler-10',
  6541920 => 'Kepler-11',
  11804465 => 'Kepler-12',
  10264660 => 'Kepler-14',
  11359879 => 'Kepler-15',
  12644769 => 'Kepler-16',
  10619192 => 'Kepler-17',
  8644288 => 'Kepler-18',
  2571238 => 'Kepler-19',
  6850504 => 'Kepler-20',
  3632418 => 'HD 179070',
  10593626 => 'Kepler-22',
  11512246 => 'Kepler-23',
  3231341 => 'Kepler-24',
  4349452 => 'Kepler-25',
  9757613 => 'Kepler-26',
  6949607 => 'Kepler-28',
  10358759 => 'Kepler-29',
  3832474 => 'Kepler-30',
  9347899 => 'Kepler-31',
  9787239 => 'Kepler-32',
  9458613 => 'Kepler-33',
  8572936 => 'Kepler-34',
  9837578 => 'Kepler-35',
  9941662 => 'KOI-13',
  5807616 => 'KOI-55',
  9818381 => 'KOI-135',
  9410930 => 'KOI-196',
  9305831 => 'KOI-204',
  5794240 => 'KOI-254',
  9478990 => 'KOI-423',
  10418224 => 'KOI-428',
  10227020 => 'KOI-730',
  7109675 => 'KOI-872',
  8561063 => 'KOI-961',
  6185331 => 'KIC 6185331',
  10905746 => 'KIC 10905746'
);

my %stars = ();
my %planets = ();

# Read in data for planet candidates
open my $candidatesfile, '<', 'koiTable_2012Feb26.txt' or die "Cannot open appendixTable_2012Feb26.txt\n";

# Skip header
for (my $i = 0; $i < 57; ++$i)
{
  my $dummy = <$candidatesfile>;
}

while (my $record = <$candidatesfile>)
{
  chomp $record;
  my $id  = trim(substr($record,   0,  7));
  my $koi = trim(substr($record,   0,  4));
  my $cdt = trim(substr($record,   5,  2));
  my $kic = trim(substr($record,   8,  9));
  my $t0  = trim(substr($record,  26, 10));
  my $per = trim(substr($record,  46, 12));
  my $rad = trim(substr($record,  70,  6));
  my $a   = trim(substr($record,  77,  6));
  my $teq = trim(substr($record,  84,  5));
  my $b   = trim(substr($record, 148,  8));
  my $tef = trim(substr($record, 179,  5));
  my $sra = trim(substr($record, 191,  6));

  if (exists($exclude{$kic}))
  {
    print "KOI $id host star in Celestia catalogue. Skipping.\n";
    next;
  }
  if ($per <= 0)
  {
    print "KOI $id bad period. Skipping.\n";
    next;
  }
  if ($rad <= 0)
  {
    print "KOI $id bad radius. Skipping.\n";
    next;
  }
  if ($a <= 0)
  {
    print "KOI $id bad semimajor axis. Skipping.\n";
    next;
  }
  if ($b < 0)
  {
    print "KOI $id bad impact parameter. Skipping.\n";
    next;
  }
  if ((SOLAR_RADIUS_AU * $sra) >= $a)
  {
    print "KOI $id semimajor axis within star. Skipping.\n";
    next;
  }

  if (!exists($stars{$kic}{koi}))
  {
    $stars{$kic} = {
      koi => $koi,
      teff => $tef,
      rad => $sra
    }
  }

  $planets{$id} = {
    koi => $koi,
    planet => $cdt,
    kic => $kic,
    t0 => $t0,
    period => $per,
    radius => $rad,
    a => $a,
    teq => $teq,
    b => $b
  }
}

close $candidatesfile;

my $kic = new IO::Uncompress::Gunzip 'kic.txt.gz' or die "IO::Uncompress:Gunzip failed: $GunzipError\n";

while (my $record = <$kic>)
{
  chomp $record;
  my ($ra, $dec, $gmag, $kic, $gr, $twomass) = (split(/\|/, $record))[0, 1, 5, 15, 38, 43];

  next if (!exists($stars{$kic}));

  if ($ra eq '')
  {
    print "KOI $stars{$kic}{koi} missing RA. Skipping.\n";
    next;
  }
  if ($dec eq '')
  {
    print "KOI $stars{$kic}{koi} missing Dec. Skipping.\n";
    next;
  }
  if ($gmag eq '')
  {
    print "KOI $stars{$kic}{koi} missing g magnitude. Skipping.\n";
    next;
  }
  if ($gr eq '')
  {
    print "KOI $stars{$kic}{koi} missing g-r index. Skipping.\n";
    next;
  }

  $twomass =~ s/\s+//g;

  $stars{$kic}{twomass} = $twomass;
  $stars{$kic}{ra} = $ra * 15;
  $stars{$kic}{dec} = $dec;
  $stars{$kic}{gmag} = $gmag;
  $stars{$kic}{gr} = $gr;
}

close $kic;

# Star temperatures in Celestia, taken from star.cpp
my %startypes = (
  'O3' => 52500, 'O4' => 48000, 'O5' => 44500, 'O6' => 41000, '07' => 38000, 'O8' => 35800, 'O9' => 33000,
  'B0' => 30000, 'B1' => 25400, 'B2' => 22000, 'B3' => 18700, 'B4' => 17000, 'B5' => 15400, 'B6' => 14000, 'B7' => 13000, 'B8' => 11900, 'B9' => 10500,
  'A0' => 9520, 'A1' => 9230, 'A2' => 8970, 'A3' => 8720, 'A4' => 8460, 'A5' => 8200, 'A6' => 8020, 'A7' => 7850, 'A8' => 7580, 'A9' => 7390,
  'F0' => 7200, 'F1' => 7050, 'F2' => 6890, 'F3' => 6740, 'F4' => 6590, 'F5' => 6440, 'F6' => 6360, 'F7' => 6280, 'F8' => 6200, 'F9' => 6110,
  'G0' => 6030, 'G1' => 5940, 'G2' => 5860, 'G3' => 5830, 'G4' => 5800, 'G5' => 5770, 'G6' => 5700, 'G7' => 5630, 'G8' => 5570, 'G9' => 5410,
  'K0' => 5250, 'K1' => 5080, 'K2' => 4900, 'K3' => 4730, 'K4' => 4590, 'K5' => 4350, 'K6' => 4200, 'K7' => 4060, 'K8' => 3990, 'K9' => 3920,
  'M0' => 3850, 'M1' => 3720, 'M2' => 3580, 'M3' => 3470, 'M4' => 3370, 'M5' => 3240, 'M6' => 3050, 'M7' => 2940, 'M8' => 2640, 'M9' => 2000
);

my $starcount = 0;
my $planetcount = 0;

# Output star data
open my $stcfile, '>', 'kepler_candidates.stc' or die "Cannot open kepler_candidates.stc for writing.\n";

# Write header
print $stcfile "# kepler_candidates.stc\n";
print $stcfile "#\n";
print $stcfile "# Extracted from the Kepler Input Catalog (KIC) available at\n";
print $stcfile "#   http://archive.stsci.edu/kepler/kepler_fov/search.php?form=fuf\n";
print $stcfile "# and the Kepler candidates list available at\n";
print $stcfile "#   http://kepler.nasa.gov/Mission/discoveries/candidates/\n";
print $stcfile "#\n";
print $stcfile "# Apparent magnitude is computed from g-band magnitudes and g-r colour index\n";
print $stcfile "# using the formulae given in Windhorst et al. 1991, ApJ, 380, 362.\n";
print $stcfile "# Distance is computed from apparent magnitude, effective temperature and\n";
print $stcfile "# stellar radius using the bolometric correction formula given at\n";
print $stcfile "#   http://www.bogan.ca/astro/stars/star_lum.html\n";
print $stcfile "# Spectral types are assigned based on closest match between the effective\n";
print $stcfile "# temperature in the catalogue and the Celestia default spectral types.\n";
print $stcfile "#\n";
print $stcfile "# Processed on $date by $author\n\n";

for my $kic (sort { $stars{$a}{koi} <=> $stars{$b}{koi} } keys %stars)
{
  next if (!exists($stars{$kic}{ra}));

  my $name = "KOI-$stars{$kic}{koi}:KIC $kic";
  $name .= ":2MASS J$stars{$kic}{twomass}" if($stars{$kic}{twomass} ne '');

  # Get radius of star in kilometres
  my $radius = $stars{$kic}{rad} * SOLAR_RADIUS_KM;

  # g to V conversion from Windhorst et al. 1991, ApJ, 380, 362
  my $Vmag = $stars{$kic}{gmag} - 0.03 - 0.42 * $stars{$kic}{gr};
  my $MBol = SOLAR_BOLO_MAG - 2.5 * log10((($stars{$kic}{teff} / SOLAR_TEMPERATURE) ** 4) * ($stars{$kic}{rad} ** 2));

  # Bolometric correction from http://www.bogan.ca/astro/stars/star_lum.html
  my $logTeff4 = log10($stars{$kic}{teff}) - 4;
  my $BC = -8.499 * ($logTeff4 ** 4) + 13.421 * ($logTeff4 ** 3) - 8.131 * ($logTeff4 ** 2) - 3.901 * $logTeff4 - 0.438;

  # Compute distance from absolute and apparent magnitude
  my $MV = $MBol - $BC;
  my $distance = PARSEC_LY * (10 ** (($Vmag - $MV) / 5 + 1));

  # Get closest Celestia star spectral type
  my $startype = '?';
  my $delta = 999999999;
  for my $type (keys %startypes)
  {
    my $newdelta = abs($startypes{$type} - $stars{$kic}{teff});
    if($newdelta < $delta)
    {
      $startype = $type;
      $delta = $newdelta;
    }
  }

  print $stcfile "\"$name\"\n";
  print $stcfile "\{\n";
  print $stcfile sprintf("\tRA %.6f\n", $stars{$kic}{ra});
  print $stcfile sprintf("\tDec %.6f\n", $stars{$kic}{dec});
  print $stcfile sprintf("\tDistance %.0f # estimate from luminosity and colour index\n", $distance);
  print $stcfile sprintf("\tAppMag %.2f # estimate from g-magnitude and colour index\n", $Vmag);
  print $stcfile "\tSpectralType \"$startype\" # estimate from temperature\n";
  print $stcfile sprintf("\tRadius %.0f\n", $radius);
  print $stcfile "\}\n\n";

  ++$starcount;
}

close $stcfile;

# Output planet data
open my $sscfile, '>', 'kepler_candidates.ssc' or die "Cannot open kepler_candidates.ssc for writing.\n";

# Write header
print $sscfile "# kepler_candidates.ssc\n";
print $sscfile "#\n";
print $sscfile "# Extracted from the Kepler Input Catalog (KIC) available at\n";
print $sscfile "#   http://archive.stsci.edu/kepler/kepler_fov/search.php?form=fuf\n";
print $sscfile "# and the Kepler candidates list available at\n";
print $sscfile "#   http://kepler.nasa.gov/Mission/discoveries/candidates/\n";
print $sscfile "#\n";
print $sscfile "# Orbits are assumed to be circular with inclination derived from the transit\n";
print $sscfile "# impact parameter, orbital semimajor axis and stellar radius. Planet textures\n";
print $sscfile "# are assigned based on the planetary radius as follows:\n";
print $sscfile "#   Earths:                   Rp <  1.25 - asteroid.*\n";
print $sscfile "#   Super-Earths:      1.25 < Rp <  2.0  - asteroid.*\n";
print $sscfile "#   Neptunes:          2.0  < Rp <  6.0  - venuslike.*\n";
print $sscfile "#   Jupiters:          6.0  < Rp < 15.0  - exo-class#.*\n";
print $sscfile "#   Super-Jupiters:   15.0  < Rp < 22.4  - exo-class#.*\n";
print $sscfile "#   Brown Dwarfs (?): 22.4  < Rp         - browndwarf.*\n";
print $sscfile "# The texture for Jupiters and Super-Jupiters is chosen based on equilibrium\n";
print $sscfile "# temperature as follows:\n";
print $sscfile "#          Teq <  150 - exo-class1.*\n";
print $sscfile "#    150 < Teq <  350 - exo-class2.*\n";
print $sscfile "#    350 < Teq <  900 - exo-class3.*\n";
print $sscfile "#    900 < Teq < 1500 - exo-class4.*\n";
print $sscfile "#   1500 < Teq        - exo-class5.*\n";
print $sscfile "# Glowing nightsides are added to planets with Teq > 1000 K.\n";
print $sscfile "#\n";
print $sscfile "# Rotation rates are estimated using the tidal spindown formulae in Goldreich\n";
print $sscfile "# and Soter (1966) with an assumed age of 4 Gyr and the following initial\n";
print $sscfile "# rotation periods and tidal quality factors:\n";
print $sscfile "#   Earths:           Q=100,    Pi = 5 hours\n";
print $sscfile "#   Super-Earths:     Q=100,    Pi = 5 hours\n";
print $sscfile "#   Neptunes:         Q=100000, Pi = 17 hours\n";
print $sscfile "#   Jupiters:         Q=100000, Pi = 10 hours\n";
print $sscfile "#   Super-Jupiters:   Q=100000, Pi = 10 hours\n";
print $sscfile "#   Brown Dwarfs (?): Q=100000, Pi = 10 hours\n";
print $sscfile "# Planetary masses are estimated using the mass-radius relationship\n";
print $sscfile "#   M/Me = (R/Re)^2.03\n";
print $sscfile "# given in Lissauer et al. (2011), ApJS 197, id. 8\n";
print $sscfile "#\n";
print $sscfile "# Processed on $date by $author\n\n";

for my $id (sort sort_koi keys %planets)
{
  next if (!exists($stars{$planets{$id}{kic}}{ra}));

  # Orbital period in years, radius in kilometres, transit epoch
  my $period = $planets{$id}{period} / EARTH_YEAR;
  my $radius = $planets{$id}{radius} * EARTH_RADIUS_KM;
  my $epoch = $planets{$id}{t0} + 2454900.0;

  # Convert impact parameter to inclination in the sky plane
  my $skyinc = rad2deg(acos($planets{$id}{b} * SOLAR_RADIUS_AU * $stars{$planets{$id}{kic}}{rad} / $planets{$id}{a}));

  # Do the orbit transformation
  my ($inc, $node, $omega) = TransformElements($stars{$planets{$id}{kic}}{ra}, $stars{$planets{$id}{kic}}{dec}, $skyinc, 0, 0);

  # At time of transit, true anomaly is 90 degrees minus omega. For circular orbits, mean anomaly = true anomaly
  my $ma = 90 - $omega;
  $ma += 360 if($ma < 0);

  # Assign textures based on Kepler radius bins
  # Earths:                 Rp <  1.25
  # Super-Earths:    1.25 < Rp <  2.0
  # Neptunes:        2.0  < Rp <  6.0
  # Jupiters:        6.0  < Rp < 15.0
  # Super-Jupiters: 15.0  < Rp < 22.4
  # Brown Dwarfs?   22.4  < Rp

  print $sscfile "\"$planets{$id}{planet}\" \"KOI-$planets{$id}{koi}\"\n";
  print $sscfile "\{\n";

  print $sscfile sprintf("\tRadius %.0f\n\n", $radius);

  my $initialRotation = 0;
  my $qualityFactor = 0;

  if ($planets{$id}{radius} < 2.0)
  {
    # Earth or Super-Earth
    print $sscfile "\tTexture \"asteroid.*\"\n";
    $initialRotation = 5;
    $qualityFactor = 100;
  }
  elsif ($planets{$id}{radius} < 6.0)
  {
    # Neptune
    print $sscfile "\tTexture \"venuslike.*\"\n";
    print $sscfile "\tNightTexture \"venuslikenight.*\"\n" if($planets{$id}{teq} >= 1000);
    $initialRotation = 17;
    $qualityFactor = 100000;
  }
  elsif ($planets{$id}{radius} < 22.4)
  {
    # Jupiter or Super-Jupiter - use Sudarsky classification system
    if ($planets{$id}{teq} < 150)
    {
      print $sscfile "\tTexture \"exo-class1.*\"\n";
      print $sscfile "\tColor [0.96 0.96 1]\n";
      print $sscfile "\tAlbedo 0.5\n";
    }
    elsif ($planets{$id}{teq} < 350)
    {
      print $sscfile "\tTexture \"exo-class2.*\"\n";
      print $sscfile "\tColor [0.98 0.97 1]\n";
      print $sscfile "\tAlbedo 0.75\n";
    }
    elsif ($planets{$id}{teq} < 900)
    {
      print $sscfile "\tTexture \"exo-class3.*\"\n";
      print $sscfile "\tColor [0.53 0.62 1]\n";
      print $sscfile "\tAlbedo 0.1\n";
    }
    elsif ($planets{$id}{teq} < 1500)
    {
      print $sscfile "\tTexture \"exo-class4.*\"\n";
      print $sscfile "\tNightTexture \"exo-class4night.*\"\n" if($planets{$id}{teq} >= 1000);
      print $sscfile "\tColor [0 0.36 1]\n";
      print $sscfile "\tAlbedo 0.05\n";
    }
    else
    {
      print $sscfile "\tTexture \"exo-class5.*\"\n";
      print $sscfile "\tNightTexture \"exo-class5night.*\"\n";
      print $sscfile "\tColor [0.49 0.89 1]\n";
      print $sscfile "\tAlbedo 0.5\n";
    }
    $initialRotation = 10;
    $qualityFactor = 100000;
  }
  else
  {
    # Brown dwarfs maybe?
    print $sscfile "\tTexture \"browndwarf.*\"\n";
    print $sscfile "\tEmissive true\n";
    $initialRotation = 10;
    $qualityFactor = 100000;
  }

  # Estimate mass of the planet from mass-radius relationship
  my $mplanet = $planets{$id}{radius} ** 2.06;

  my $finalRotation = TidalSpindown(
    $initialRotation,
    $qualityFactor,
    $planets{$id}{radius},
    $mplanet,
    $planets{$id}{a},
    $planets{$id}{period});

  $finalRotation = int($finalRotation); # round down to avoid slower rotation than orbit

  print $sscfile sprintf("\n\tMass %.0f # estimate from mass-radius relationship\n\n", $mplanet);

  print $sscfile "\tEllipticalOrbit \{\n";
  print $sscfile sprintf("\t\tEpoch %.5f # transit time\n", $epoch);
  print $sscfile sprintf("\t\tPeriod %.9f # %.7f days\n", $period, $planets{$id}{period});
  print $sscfile sprintf("\t\tSemiMajorAxis %.3f\n", $planets{$id}{a});
  print $sscfile sprintf("\t\tInclination %.2f\n", $inc);
  print $sscfile sprintf("\t\tAscendingNode %.2f\n", $node);
  print $sscfile sprintf("\t\tMeanAnomaly %.2f\n", $ma);
  print $sscfile "\t\}\n\n";
  print $sscfile "\tUniformRotation \{\n";
  if($finalRotation > 0)
  {
    print $sscfile "\t\tPeriod $finalRotation # estimate from spindown model\n";
  }
  else
  {
    print $sscfile "\t\t# likely tidally-locked\n";
  }
  print $sscfile sprintf("\t\tInclination %.2f # to match orbit\n", $inc);
  print $sscfile sprintf("\t\tAscendingNode %.2f # to match orbit\n", $node);
  print $sscfile "\t\}\n";
  print $sscfile "\}\n\n";
  ++$planetcount
}

close $sscfile;

# Output statistics.
print "Output $starcount stars.\n";
print "Output $planetcount planets.\n";

# End of program

# Trim whitespace from ends of a string
sub trim()
{
  my $string = $_[0];
  $string =~ s/^\s+//;
  $string =~ s/\s+$//;
  return $string;
}

# Sort by increasing KOI number and increasing semimajor axis
sub sort_koi()
{
  my $koicmp = $planets{$a}{koi} <=> $planets{$b}{koi};
  return ($koicmp != 0) ? $koicmp : ($planets{$a}{a} <=> $planets{$b}{a});
}

# Base-10 logarithm
sub log10 {
  my $n = $_[0];
  return log($_[0]) / log(10);
}

# Orbital elements transformation
sub TransformElements
{
  my ($RA, $Dec, $inc, $OMEGA, $omega) = @_;

  $RA  = deg2rad($RA);
  $Dec = deg2rad($Dec);

  # Calculate quaternion in sky-plane frame
  $inc   = deg2rad($inc);
  $OMEGA = deg2rad($OMEGA);
  $omega = deg2rad($omega);

  my $t = Math::Quaternion::rotation($OMEGA, 0, 0, 1)
    * Math::Quaternion::rotation($inc  , 1, 0, 0)
    * Math::Quaternion::rotation($omega, 0, 0, 1);

  # Convert to ecliptic frame
  my $r = Math::Quaternion::rotation(deg2rad(- EARTH_OBLIQUITY), 1, 0, 0)
    * Math::Quaternion::rotation($RA, 0, 0, 1)
    * Math::Quaternion::rotation(-$Dec - pi/2, 0, 1, 0)
    * $t;

  # Calculate transformed orbital elements
  my @q = @{$r};

  my $calcInc;
  my $calcOmega;
  my $calcOMEGA;

  if(($q[1] == 0) && ($q[2] == 0))
  {
    # Special case: inclination = 0 degrees
    $calcInc = 0;
    $calcOmega = rad2deg(atan2($q[3], $q[0]));
    $calcOMEGA = 0;
  }
  elsif(($q[0] == 0) && ($q[3] == 0))
  {
    # Special case: inclination = 180 degrees
    $calcInc = 180;
    $calcOmega = rad2deg(atan2(-$q[2], $q[1]));
    $calcOMEGA = 0;
  }
  else
  {
    # General case
    $calcInc = rad2deg(acos($q[0] * $q[0] - $q[1] * $q[1] - $q[2] * $q[2] + $q[3] * $q[3]));
    $calcOmega = rad2deg(atan2($q[1] * $q[3] - $q[0] * $q[2], $q[0] * $q[1] + $q[2] * $q[3]));
    $calcOMEGA = rad2deg(atan2($q[1] * $q[3] + $q[0] * $q[2], $q[0] * $q[1] - $q[2] * $q[3]));
  }

  $calcOmega += 360 if($calcOmega < 0);
  $calcOMEGA += 360 if($calcOMEGA < 0);

  return ($calcInc, $calcOMEGA, $calcOmega);
}

sub TidalSpindown
{
  # Rough estimate of tidal spindown
  my ($initialRotation, $Q, $r, $mplanet, $a, $period) = @_;

  my $radius = $r * EARTH_RADIUS_KM;
  my $distance = $a * AU_KM;
  my $omega1 = 2 * pi / ($initialRotation / 24);

  # Compute mass of the star from period and semimajor axis
  my $mstar = 4 * (pi ** 2) * ($distance ** 3) / (GRAVITATION * ($period ** 2));

  # Calculate spindown rate using Goldreich and Soter (1966)
  my $rate = 45 / 8 * GRAVITATION * ($mstar ** 2) * ($radius ** 3) / ($mplanet * $Q * ($distance ** 6));
  my $delta = $rate * 4e+9 * EARTH_YEAR;
  my $omega2 = $omega1 - $delta;

  my $finalRotation = 2 * pi / ($omega2);
  if(($finalRotation < 0) || ($finalRotation > $period))
  {
    return 0;
  }
  else
  {
    return $finalRotation * 24
  }
}

sub download_file {

  my ($ua, $url, $dest) = @_;

  my $result;

  $result = $ua->head($url);
  my $headers = $result->headers;
  my $totalsize = $headers->content_length;

  my $progress = Term::ProgressBar->new({count => $totalsize, ETA => 'linear'});
  $progress->minor(0);
  $progress->max_update_rate(1);

  my $nextupdate = 0;
  my $size = 0;

  open my $outfile, '>', $dest or die "Cannot open $dest for writing.\n";
  binmode $outfile;

  my $req = HTTP::Request->new(GET => $url);
  $result = $ua->request($req,
    sub {
      my ($data, $response, $protocol) = @_;
      print $outfile $data;
      $size += length($data);
      $nextupdate = $progress->update($size) if ($size > $nextupdate);
    }
  );
  $progress->update($totalsize) if ($totalsize > $nextupdate);

  print "\n";

  close $outfile;

  if ($result->is_success)
  {
    print "Download successful.\n";
    return 1;
  }
  else
  {
    print "Download error: ", $result->status_line, "\n";
    return 0;
  }
}